@@ -15,7 +15,7 @@
 //! can instead just be assumed to not exist:
 //!
 //! - If you try to search for a user with a null character in their name,
-//!   such a user could not exist anyway—so it’s OK to return `None`.
+//!   such a user could not exist anyway — so it’s OK to return `None`.
 //! - If the OS returns user information with a null character in a field,
 //!   then that field will just be truncated instead, which is valid behaviour
 //!   for a `CString`.
@@ -31,47 +31,26 @@
 
 use std::ffi::{CStr, CString, OsStr, OsString};
 use std::fmt;
+use std::mem;
 use std::io::{Result as IoResult, Error as IoError};
 use std::os::unix::ffi::OsStrExt;
-use std::ptr::read;
+use std::ptr;
 use std::sync::Arc;
 
-use libc::{
-    c_char,
-    uid_t,
-    gid_t,
-    getpwuid,
-    getpwnam,
-    getgrgid,
-    getgrnam,
-    getuid,
-    geteuid,
-    getgid,
-    getegid,
-    getgroups,
-    getgrouplist,
-    c_int,
-};
-#[cfg(not(target_os = "android"))]
-use libc::{
-    setpwent,
-    getpwent,
-    endpwent
-};
-
+use libc::{c_char, uid_t, gid_t, c_int};
 use libc::passwd as c_passwd;
 use libc::group as c_group;
 
+
 /// Information about a particular user.
+///
+/// For more information, see the [module documentation](index.html).
 #[derive(Clone)]
 pub struct User {
     uid: uid_t,
     primary_group: gid_t,
     extras: os::UserExtras,
-
-    /// This user’s name, as an owned `OsString` possibly shared with a cache.
-    /// Prefer using the `name()` accessor to using this field, if possible.
-    pub name_arc: Arc<OsString>,
+    pub(crate) name_arc: Arc<OsString>,
 }
 
 impl User {
@@ -79,31 +58,83 @@ impl User {
     /// Create a new `User` with the given user ID, name, and primary
     /// group ID, with the rest of the fields filled with dummy values.
     ///
-    /// This method does not actually create a new user on the system—it
+    /// This method does not actually create a new user on the system — it
     /// should only be used for comparing users in tests.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use users::User;
+    ///
+    /// let user = User::new(501, "stevedore", 100);
+    /// ```
     pub fn new<S: AsRef<OsStr> + ?Sized>(uid: uid_t, name: &S, primary_group: gid_t) -> Self {
         let name_arc = Arc::new(name.into());
         let extras = os::UserExtras::default();
-    
+
         User { uid, name_arc, primary_group, extras }
     }
 
     /// Returns this user’s ID.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use users::User;
+    ///
+    /// let user = User::new(501, "stevedore", 100);
+    /// assert_eq!(user.uid(), 501);
+    /// ```
     pub fn uid(&self) -> uid_t {
         self.uid
     }
 
     /// Returns this user’s name.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use std::ffi::OsStr;
+    /// use users::User;
+    ///
+    /// let user = User::new(501, "stevedore", 100);
+    /// assert_eq!(user.name(), OsStr::new("stevedore"));
+    /// ```
     pub fn name(&self) -> &OsStr {
         &**self.name_arc
     }
 
     /// Returns the ID of this user’s primary group.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use users::User;
+    ///
+    /// let user = User::new(501, "stevedore", 100);
+    /// assert_eq!(user.primary_group_id(), 100);
+    /// ```
     pub fn primary_group_id(&self) -> gid_t {
         self.primary_group
     }
 
-    /// Returns a list of groups this user is a member of
+    /// Returns a list of groups this user is a member of. This involves
+    /// loading the groups list, as it is _not_ contained within this type.
+    ///
+    /// # libc functions used
+    ///
+    /// - [`getgrouplist`](https://docs.rs/libc/*/libc/fn.getgrouplist.html)
+    ///
+    /// # Examples
+    ///
+    /// ```no_run
+    /// use users::User;
+    ///
+    /// let user = User::new(501, "stevedore", 100);
+    /// for group in user.groups().expect("User not found") {
+    ///     println!("User is in group: {:?}", group.name());
+    /// }
+    /// ```
     pub fn groups(&self) -> Option<Vec<Group>> {
         get_user_groups(self.name(), self.primary_group_id())
     }
@@ -127,14 +158,13 @@ impl fmt::Debug for User {
 
 
 /// Information about a particular group.
+///
+/// For more information, see the [module documentation](index.html).
 #[derive(Clone)]
 pub struct Group {
     gid: gid_t,
     extras: os::GroupExtras,
-
-    /// This group’s name, as an owned `OsString` possibly shared with a cache.
-    /// Prefer using the `name()` accessor to using this field, if possible.
-    pub name_arc: Arc<OsString>,
+    pub(crate) name_arc: Arc<OsString>,
 }
 
 impl Group {
@@ -142,21 +172,48 @@ impl Group {
     /// Create a new `Group` with the given group ID and name, with the
     /// rest of the fields filled in with dummy values.
     ///
-    /// This method does not actually create a new group on the system—it
+    /// This method does not actually create a new group on the system — it
     /// should only be used for comparing groups in tests.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use users::Group;
+    ///
+    /// let group = Group::new(102, "database");
+    /// ```
     pub fn new<S: AsRef<OsStr> + ?Sized>(gid: gid_t, name: &S) -> Self {
         let name_arc = Arc::new(name.into());
         let extras = os::GroupExtras::default();
-    
+
         Group { gid, name_arc, extras }
     }
 
     /// Returns this group’s ID.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use users::Group;
+    ///
+    /// let group = Group::new(102, "database");
+    /// assert_eq!(group.gid(), 102);
+    /// ```
     pub fn gid(&self) -> gid_t {
         self.gid
     }
 
     /// Returns this group’s name.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use std::ffi::OsStr;
+    /// use users::Group;
+    ///
+    /// let group = Group::new(102, "database");
+    /// assert_eq!(group.name(), OsStr::new("database"));
+    /// ```
     pub fn name(&self) -> &OsStr {
         &**self.name_arc
     }
@@ -188,7 +245,7 @@ unsafe fn from_raw_buf(p: *const c_char) -> OsString {
 }
 
 unsafe fn passwd_to_user(pointer: *const c_passwd) -> Option<User> {
-    if let Some(passwd) = pointer.as_ref().map(|p| read(p)) {
+    if let Some(passwd) = pointer.as_ref().map(|p| ptr::read(p)) {
         let name = Arc::new(from_raw_buf(passwd.pw_name));
 
         Some(User {
@@ -204,7 +261,7 @@ unsafe fn passwd_to_user(pointer: *const c_passwd) -> Option<User> {
 }
 
 unsafe fn struct_to_group(pointer: *const c_group) -> Option<Group> {
-    if let Some(group) = pointer.as_ref().map(|p| read(p)) {
+    if let Some(group) = pointer.as_ref().map(|p| ptr::read(p)) {
         let name = Arc::new(from_raw_buf(group.gr_name));
 
         Some(Group {
@@ -245,106 +302,361 @@ unsafe fn members(groups: *mut *mut c_char) -> Vec<OsString> {
 
 /// Searches for a `User` with the given ID in the system’s user database.
 /// Returns it if one is found, otherwise returns `None`.
+///
+/// # libc functions used
+///
+/// - [`getpwuid_r`](https://docs.rs/libc/*/libc/fn.getpwuid_r.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_user_by_uid;
+///
+/// match get_user_by_uid(501) {
+///     Some(user) => println!("Found user {:?}", user.name()),
+///     None       => println!("User not found"),
+/// }
+/// ```
 pub fn get_user_by_uid(uid: uid_t) -> Option<User> {
+    let mut passwd = unsafe { mem::zeroed::<c_passwd>() };
+    let mut buf = vec![0; 2048];  // TODO: Retry with larger buffer sizes
+    let mut result = ptr::null_mut::<c_passwd>();
+
     unsafe {
-        let passwd = getpwuid(uid);
-        passwd_to_user(passwd)
+        libc::getpwuid_r(uid, &mut passwd, buf.as_mut_ptr(), buf.len(), &mut result);
     }
+
+    if result.is_null() {
+        // There is no such user, or an error has occurred.
+        // errno gets set if there’s an error.
+        return None;
+    }
+
+    if result != &mut passwd {
+        // The result of getpwuid_r should be its input passwd.
+        return None;
+    }
+
+    unsafe { passwd_to_user(result) }
 }
 
 /// Searches for a `User` with the given username in the system’s user database.
 /// Returns it if one is found, otherwise returns `None`.
+///
+/// # libc functions used
+///
+/// - [`getpwnam_r`](https://docs.rs/libc/*/libc/fn.getpwnam_r.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_user_by_name;
+///
+/// match get_user_by_name("stevedore") {
+///     Some(user) => println!("Found user #{}", user.uid()),
+///     None       => println!("User not found"),
+/// }
+/// ```
 pub fn get_user_by_name<S: AsRef<OsStr> + ?Sized>(username: &S) -> Option<User> {
-    if let Ok(username) = CString::new(username.as_ref().as_bytes()) {
-        unsafe {
-            let passwd = getpwnam(username.as_ptr());
-            passwd_to_user(passwd)
+    let username = match CString::new(username.as_ref().as_bytes()) {
+        Ok(u)  => u,
+        Err(_) => {
+            // The username that was passed in contained a null character,
+            // which will match no usernames.
+            return None;
         }
+    };
+
+    let mut passwd = unsafe { mem::zeroed::<c_passwd>() };
+    let mut buf = vec![0; 2048];  // TODO: Retry with larger buffer sizes
+    let mut result = ptr::null_mut::<c_passwd>();
+
+    unsafe {
+        libc::getpwnam_r(username.as_ptr(), &mut passwd, buf.as_mut_ptr(), buf.len(), &mut result);
     }
-    else {
-        // The username that was passed in contained a null character.
-        // This will *never* find anything, so just return `None`.
-        // (I can’t figure out a pleasant way to signal an error here)
-        None
+
+    if result.is_null() {
+        // There is no such user, or an error has occurred.
+        // errno gets set if there’s an error.
+        return None;
     }
+
+    if result != &mut passwd {
+        // The result of getpwnam_r should be its input passwd.
+        return None;
+    }
+
+    unsafe { passwd_to_user(result) }
 }
 
 /// Searches for a `Group` with the given ID in the system’s group database.
 /// Returns it if one is found, otherwise returns `None`.
+///
+/// # libc functions used
+///
+/// - [`getgrgid_r`](https://docs.rs/libc/*/libc/fn.getgrgid_r.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_group_by_gid;
+///
+/// match get_group_by_gid(102) {
+///     Some(group) => println!("Found group {:?}", group.name()),
+///     None        => println!("Group not found"),
+/// }
+/// ```
 pub fn get_group_by_gid(gid: gid_t) -> Option<Group> {
+    let mut passwd = unsafe { mem::zeroed::<c_group>() };
+    let mut buf = vec![0; 2048];  // TODO: Retry with larger buffer sizes
+    let mut result = ptr::null_mut::<c_group>();
+
     unsafe {
-        let group = getgrgid(gid);
-        struct_to_group(group)
+        libc::getgrgid_r(gid, &mut passwd, buf.as_mut_ptr(), buf.len(), &mut result);
     }
+
+    if result.is_null() {
+        // There is no such group, or an error has occurred.
+        // errno gets set if there’s an error.
+        return None;
+    }
+
+    if result != &mut passwd {
+        // The result of getgrgid_r should be its input struct.
+        return None;
+    }
+
+    unsafe { struct_to_group(result) }
 }
 
 /// Searches for a `Group` with the given group name in the system’s group database.
 /// Returns it if one is found, otherwise returns `None`.
-pub fn get_group_by_name<S: AsRef<OsStr> + ?Sized>(group_name: &S) -> Option<Group> {
-    if let Ok(group_name) = CString::new(group_name.as_ref().as_bytes()) {
-        unsafe {
-            let group = getgrnam(group_name.as_ptr());
-            struct_to_group(group)
+///
+/// # libc functions used
+///
+/// - [`getgrnam_r`](https://docs.rs/libc/*/libc/fn.getgrnam_r.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_group_by_name;
+///
+/// match get_group_by_name("db-access") {
+///     Some(group) => println!("Found group #{}", group.gid()),
+///     None        => println!("Group not found"),
+/// }
+/// ```
+pub fn get_group_by_name<S: AsRef<OsStr> + ?Sized>(groupname: &S) -> Option<Group> {
+    let groupname = match CString::new(groupname.as_ref().as_bytes()) {
+        Ok(u)  => u,
+        Err(_) => {
+            // The groupname that was passed in contained a null character,
+            // which will match no usernames.
+            return None;
         }
+    };
+
+    let mut group = unsafe { mem::zeroed::<c_group>() };
+    let mut buf = vec![0; 2048];  // TODO: Retry with larger buffer sizes
+    let mut result = ptr::null_mut::<c_group>();
+
+    unsafe {
+        libc::getgrnam_r(groupname.as_ptr(), &mut group, buf.as_mut_ptr(), buf.len(), &mut result);
     }
-    else {
-        // The group name that was passed in contained a null character.
-        // This will *never* find anything, so just return `None`.
-        // (I can’t figure out a pleasant way to signal an error here)
-        None
+
+    if result.is_null() {
+        // There is no such group, or an error has occurred.
+        // errno gets set if there’s an error.
+        return None;
     }
+
+    if result != &mut group {
+        // The result of getgrnam_r should be its input struct.
+        return None;
+    }
+
+    unsafe { struct_to_group(result) }
 }
 
 /// Returns the user ID for the user running the process.
+///
+/// # libc functions used
+///
+/// - [`getuid`](https://docs.rs/libc/*/libc/fn.getuid.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_current_uid;
+///
+/// println!("The ID of the current user is {}", get_current_uid());
+/// ```
 pub fn get_current_uid() -> uid_t {
-    unsafe { getuid() }
+    unsafe { libc::getuid() }
 }
 
 /// Returns the username of the user running the process.
+///
+/// This function to return `None` if the current user does not exist, which
+/// could happed if they were deleted after the program started running.
+///
+/// # libc functions used
+///
+/// - [`getuid`](https://docs.rs/libc/*/libc/fn.getuid.html)
+/// - [`getpwuid_r`](https://docs.rs/libc/*/libc/fn.getpwuid_r.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_current_username;
+///
+/// match get_current_username() {
+///     Some(uname) => println!("Running as user with name {:?}", uname),
+///     None        => println!("The current user does not exist!"),
+/// }
+/// ```
 pub fn get_current_username() -> Option<OsString> {
     let uid = get_current_uid();
     get_user_by_uid(uid).map(|u| Arc::try_unwrap(u.name_arc).unwrap())
 }
 
 /// Returns the user ID for the effective user running the process.
+///
+/// # libc functions used
+///
+/// - [`geteuid`](https://docs.rs/libc/*/libc/fn.geteuid.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_effective_uid;
+///
+/// println!("The ID of the effective user is {}", get_effective_uid());
+/// ```
 pub fn get_effective_uid() -> uid_t {
-    unsafe { geteuid() }
+    unsafe { libc::geteuid() }
 }
 
 /// Returns the username of the effective user running the process.
+///
+/// # libc functions used
+///
+/// - [`geteuid`](https://docs.rs/libc/*/libc/fn.geteuid.html)
+/// - [`getpwuid_r`](https://docs.rs/libc/*/libc/fn.getpwuid_r.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_effective_username;
+///
+/// match get_effective_username() {
+///     Some(uname) => println!("Running as effective user with name {:?}", uname),
+///     None        => println!("The effective user does not exist!"),
+/// }
+/// ```
 pub fn get_effective_username() -> Option<OsString> {
     let uid = get_effective_uid();
     get_user_by_uid(uid).map(|u| Arc::try_unwrap(u.name_arc).unwrap())
 }
 
 /// Returns the group ID for the user running the process.
+///
+/// # libc functions used
+///
+/// - [`getgid`](https://docs.rs/libc/*/libc/fn.getgid.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_current_gid;
+///
+/// println!("The ID of the current group is {}", get_current_gid());
+/// ```
 pub fn get_current_gid() -> gid_t {
-    unsafe { getgid() }
+    unsafe { libc::getgid() }
 }
 
 /// Returns the groupname of the user running the process.
+///
+/// # libc functions used
+///
+/// - [`getgid`](https://docs.rs/libc/*/libc/fn.getgid.html)
+/// - [`getgrgid`](https://docs.rs/libc/*/libc/fn.getgrgid.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_current_groupname;
+///
+/// match get_current_groupname() {
+///     Some(gname) => println!("Running as group with name {:?}", gname),
+///     None        => println!("The current group does not exist!"),
+/// }
+/// ```
 pub fn get_current_groupname() -> Option<OsString> {
     let gid = get_current_gid();
     get_group_by_gid(gid).map(|g| Arc::try_unwrap(g.name_arc).unwrap())
 }
 
 /// Returns the group ID for the effective user running the process.
+///
+/// # libc functions used
+///
+/// - [`getegid`](https://docs.rs/libc/*/libc/fn.getegid.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_effective_gid;
+///
+/// println!("The ID of the effective group is {}", get_effective_gid());
+/// ```
 pub fn get_effective_gid() -> gid_t {
-    unsafe { getegid() }
+    unsafe { libc::getegid() }
 }
 
 /// Returns the groupname of the effective user running the process.
+///
+/// # libc functions used
+///
+/// - [`getegid`](https://docs.rs/libc/*/libc/fn.getegid.html)
+/// - [`getgrgid`](https://docs.rs/libc/*/libc/fn.getgrgid.html)
+///
+/// # Examples
+///
+/// ```
+/// use users::get_effective_groupname;
+///
+/// match get_effective_groupname() {
+///     Some(gname) => println!("Running as effective group with name {:?}", gname),
+///     None        => println!("The effective group does not exist!"),
+/// }
+/// ```
 pub fn get_effective_groupname() -> Option<OsString> {
     let gid = get_effective_gid();
     get_group_by_gid(gid).map(|g| Arc::try_unwrap(g.name_arc).unwrap())
 }
 
 /// Returns the group access list for the current process.
+///
+/// # libc functions used
+///
+/// - [`getgroups`](https://docs.rs/libc/*/libc/fn.getgroups.html)
+///
+/// # Examples
+///
+/// ```no_run
+/// use users::group_access_list;
+///
+/// for group in group_access_list().expect("Error looking up groups") {
+///     println!("Process can access group #{} ({:?})", group.gid(), group.name());
+/// }
+/// ```
 pub fn group_access_list() -> IoResult<Vec<Group>> {
     let mut buff: Vec<gid_t> = vec![0; 1024];
-    
+
     let res = unsafe {
-        getgroups(1024, buff.as_mut_ptr())
+        libc::getgroups(1024, buff.as_mut_ptr())
     };
 
     if res < 0 {
@@ -359,7 +671,21 @@ pub fn group_access_list() -> IoResult<Vec<Group>> {
     }
 }
 
-/// Returns groups for a provided user name and primary group id
+/// Returns groups for a provided user name and primary group id.
+///
+/// # libc functions used
+///
+/// - [`getgrouplist`](https://docs.rs/libc/*/libc/fn.getgrouplist.html)
+///
+/// # Examples
+///
+/// ```no_run
+/// use users::get_user_groups;
+///
+/// for group in get_user_groups("stevedore", 1001).expect("Error looking up groups") {
+///     println!("User is a member of group #{} ({:?})", group.gid(), group.name());
+/// }
+/// ```
 pub fn get_user_groups<S: AsRef<OsStr> + ?Sized>(username: &S, gid: gid_t) -> Option<Vec<Group>> {
     // MacOS uses i32 instead of gid_t in getgrouplist for unknown reasons
     #[cfg(all(unix, target_os="macos"))]
@@ -373,12 +699,12 @@ pub fn get_user_groups<S: AsRef<OsStr> + ?Sized>(username: &S, gid: gid_t) -> Op
     // MacOS uses i32 instead of gid_t in getgrouplist for unknown reasons
     #[cfg(all(unix, target_os="macos"))]
     let res = unsafe {
-        getgrouplist(name.as_ptr(), gid as i32, buff.as_mut_ptr(), &mut count)
+        libc::getgrouplist(name.as_ptr(), gid as i32, buff.as_mut_ptr(), &mut count)
     };
-    
+
     #[cfg(all(unix, not(target_os="macos")))]
     let res = unsafe {
-        getgrouplist(name.as_ptr(), gid, buff.as_mut_ptr(), &mut count)
+        libc::getgrouplist(name.as_ptr(), gid, buff.as_mut_ptr(), &mut count)
     };
 
     if res < 0 {
@@ -400,7 +726,13 @@ struct AllUsers;
 
 /// Creates a new iterator over every user present on the system.
 ///
-/// ## Unsafety
+/// # libc functions used
+///
+/// - [`getpwent`](https://docs.rs/libc/*/libc/fn.getpwent.html)
+/// - [`setpwent`](https://docs.rs/libc/*/libc/fn.setpwent.html)
+/// - [`endpwent`](https://docs.rs/libc/*/libc/fn.endpwent.html)
+///
+/// # Unsafety
 ///
 /// This constructor is marked as `unsafe`, which is odd for a crate
 /// that’s meant to be a safe interface. It *has* to be unsafe because
@@ -418,16 +750,27 @@ struct AllUsers;
 /// So to iterate all users, construct the iterator inside an `unsafe`
 /// block, then make sure to not make a new instance of it until
 /// iteration is over.
+///
+/// # Examples
+///
+/// ```
+/// use users::all_users;
+///
+/// let iter = unsafe { all_users() };
+/// for user in iter {
+///     println!("User #{} ({:?})", user.uid(), user.name());
+/// }
+/// ```
 pub unsafe fn all_users() -> impl Iterator<Item=User> {
     #[cfg(not(target_os = "android"))]
-    setpwent();
+    libc::setpwent();
     AllUsers
 }
 
 impl Drop for AllUsers {
     fn drop(&mut self) {
         #[cfg(not(target_os = "android"))]
-        unsafe { endpwent() };
+        unsafe { libc::endpwent() };
     }
 }
 
@@ -440,7 +783,7 @@ impl Iterator for AllUsers {
     }
     #[cfg(not(target_os = "android"))]
     fn next(&mut self) -> Option<User> {
-        unsafe { passwd_to_user(getpwent()) }
+        unsafe { passwd_to_user(libc::getpwent()) }
     }
 }
 
